/*
 * Some code here was developed by Hendrik Sattler, <post@hendrik-sattler.de>
 * I've taken this code snippets from his program "scmxx". Thanks Hendrik,
 * for this program.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#ifndef WINDBLOWS
   #include <termios.h>
#endif
#include <unistd.h>
#include "phone.h"
#include "language.h"

#define MANUFACTURER	"AT+CGMI"
#define MODEL		"AT+CGMM"
#define PHONESTATUS	"AT+CPAS"
#define RVCALENTRY	"AT^SBNR=\"vcs\","
#define WVCALENTRY	"AT^SBNW=\"vcs\","
#define VCALTEST	"AT^SBNW=?"
#define ECHOOFF		"ATE0"

char PIN[21];
int timeout = 10;
int bvon, bbis;

char *ReadPort (int fd, char *buf, char *command, int len);
int WritePort (int fd, const char *data, size_t count);
char *tty_read (int fd, char *at_command_p);
char *tty_readline (int fd);
char *HTrim (char *str);
#ifdef WINDBLOWS
unsigned long tty_speed (int newttyspeed);
#else
speed_t tty_speed (int newttyspeed);
#endif

#ifdef WINDBLOWS
    #define index(x,y)  strchr(x,y)
#endif
int OpenHandy (char *device, int baudrate)
{
int fd;
#ifndef WINDBLOWS
struct termios newtio;
#endif
	lprintf (10068, "Opening port <%s> ... ", device);
	memset (&PIN[0], 0, sizeof (PIN));

#ifdef WINDBLOWS
	if ((fd = open (device, O_RDWR)) == -1)
#else
	if ((fd = open (device, O_RDWR | O_NOCTTY | O_NONBLOCK)) == -1)
#endif
	{
	   lfprintf (10069, stderr, "\nError opening COM-ports <%s>! (\n",
	   	device, strerror (errno));
	   return -1;
	}
#ifndef WINDBLOWS
	if (fcntl (fd, F_SETFL, 0) == -1)
	{
	   lfprintf (10070, stderr, "\nError setting attributes! (%s)\n", strerror (errno));
	   close (fd);
	   return -1;
	}

	if (tcgetattr(fd, &newtio) < 0)
	{
	   lfprintf (10071, stderr, "\nError in getting device attributes: %s\n", strerror (errno));
	   close (fd);
	   return -1;
	}

	if (cfsetispeed (&newtio, tty_speed (baudrate)) == -1) 	//input at baudrate
	{
	   lfprintf (10072, stderr, "\nError setting input speed on port! %s\n", strerror (errno));
	   close (fd);
	   return -1;
	}

	if (cfsetospeed (&newtio, tty_speed (baudrate)) == -1) 	//ouput at baudrate
	{
	   lfprintf (10073, stderr, "\nError setting output speed on port! %s\n", strerror (errno));
	   close (fd);
	   return -1;
	}
#ifndef __CYGWIN__
	cfmakeraw (&newtio); 			//make raw, 8N1
#endif

	/* CLOCAL: ignore modem control lines
	 * CREAD: enable receiver
	 * ~CSTOPB: do not send two stop bits (but one)
	 */
	newtio.c_cflag |= (CLOCAL | CREAD);
	newtio.c_cflag &= ~CSTOPB;
	newtio.c_cc[VMIN] = 0;		//be comfortable with returning this amount of characters
	newtio.c_cc[VTIME] = 10 * timeout; //try reading for this amount of time (in deciseconds)

	tcflush (fd, TCIOFLUSH);	//clear serial tty

	if (tcsetattr (fd, TCSANOW, &newtio) < 0)
	{
	   lfprintf (10074, stderr, "\nError in setting device attributes: %s\n", strerror (errno));
	   close (fd);
	   return -1;
	}
#endif    // WINDBLOWS
	printf ("OK\n");
	return fd;
}

/*
 * Diese Funktion initialisiert das Handy und ermittelt wieviele
 * Speicherplaetze fuer Kalendereintraege zur Verfuegung stehen.
 */
int InitHandy (int fd, int *rfrom, int *rto)
{
char command[31], hv0[128], hv1[128];
char *p;
char von[11], bis[11];
int i;

	*rfrom = *rto = 0;
	strcpy (command, ECHOOFF);
	sprintf (hv0, "%s\r\n", ECHOOFF);
	WritePort (fd, hv0, strlen (hv0));
	ReadPort (fd, &hv0[0], command, sizeof (hv0));

	if (strcmp (hv0, "OK"))
	{
	   fprintf (stderr, "Nicht erwartete Antwort: %s\n", hv0);
	   fprintf (stderr, "Mglicherweise ist kein Handy oder Modem angeschlossen?\n");
	   return -2;
	}

	strcpy (command, VCALTEST);
	sprintf (hv0, "%s\r\n", VCALTEST);
	WritePort (fd, hv0, strlen (hv0));
	ReadPort (fd, &hv0[0], command, sizeof (hv0));

	if (!strcmp (hv0, "ERROR") || !strncmp (hv0, "+CME ERROR", 10))
	{
	   fprintf (stderr, "Es ist kein Handy angeschlossen, oder es untersttzt keinen Kalender!\n");
	   return -1;
	}

	ReadPort (fd, &hv1[0], command, sizeof (hv1));
	
	if (strcmp (hv1, "OK"))
	{
	   fprintf (stderr, "Unbekannter Fehler: %s\n", hv1);
	   return -1;
	}

	if ((p = strstr (hv0, "vcs")) == NULL)
	{
	   fprintf (stderr, "Dieses Handy untersttzt keinen Kalender!\n");
	   return -4;
	}

	if ((p = strchr (p, '(')) == NULL)
	{
	   fprintf (stderr, "Antwort war fehlerhaft!\n");
	   return -1;
	}

	p++;
	i = 0;

	while (isdigit (*p) && i < 10)
	{
	   von[i] = *p;
	   i++;
	   p++;
	}

	von[i] = 0;
	p++;
	i = 0;

	while (isdigit (*p) && i < 10)
	{
	   bis[i] = *p;
	   i++;
	   p++;
	}

	bis[i] = 0;
	*rfrom = atoi (von);
	*rto = atoi (bis);

	/*
	 * Das SIEMENS S45 funktioniert nicht richtig, wenn man auf
	 * den Slot 0 schreibt. Es bringt die Daten auf diesem Slot
	 * durcheinander und das Handy selbst wird instabil.
	 * Deshalb setzen wir den ersten Slot auf 1 fuer den Fall das
	 * wir als ersten den Slot 0 erhalten.
	 * Dieses Handy muss ohnehin nach jedem Datentransfer
	 * abgeschalten und wieder eingeschalten werden, da es sonst
	 * beim Zugriff auf den Kalender abstuerzt!
	 */
	if (*rfrom == 0)
	   *rfrom = 1;

	bvon = *rfrom;
	bbis = *rto;

	if (Verbose)
	   printf ("Kalenderslots von %d bis %d.\n", bvon, bbis);

	return bbis;
}

void SetPin (char *p)
{
	if (!p || !strlen (p))
	{
	   PIN[0] = 0;
	   return;
	}

	strncpy (PIN, p, sizeof (PIN));
	PIN[sizeof(PIN)] = 0;
}

char *GetManufacturer (int fd, char *ret)
{
char hv0[128];

	*ret = 0;
	sprintf (hv0, "%s\r\n", MANUFACTURER);
	WritePort (fd, hv0, strlen (hv0));
	memset (&hv0[0], 0, sizeof (hv0));
	ReadPort (fd, &hv0[0], MANUFACTURER, sizeof (hv0));

	if (hv0[0])
	   strcpy (ret, hv0);

	if (strcmp (hv0, "ERROR"))
	   ReadPort (fd, &hv0[0], MANUFACTURER, sizeof (hv0));

	return ret;
}

char *GetModel (int fd, char *ret)
{
char hv0[128];

	*ret = 0;
	sprintf (hv0, "%s\r\n", MODEL);
	WritePort (fd, hv0, strlen (hv0));
	memset (&hv0[0], 0, sizeof (hv0));
	ReadPort (fd, &hv0[0], MODEL, sizeof (hv0));

	if (hv0[0])
	   strcpy (ret, hv0);

	if (strcmp (hv0, "ERROR"))
	   ReadPort (fd, &hv0[0], MODEL, sizeof (hv0));

	return ret;
}

int GetPhoneStatus (int fd)
{
char hv0[128], hv1[128], *p;
int i;

	sprintf (hv0, "%s\r\n", PHONESTATUS);
	WritePort (fd, hv0, strlen (hv0));
	memset (&hv0[0], 0, sizeof (hv0));
	ReadPort (fd, &hv0[0], PHONESTATUS, sizeof (hv0));

	if (strcmp (hv0, "ERROR") && strncmp (hv0, "+CME ERROR", 10))
	{
	   ReadPort (fd, &hv1[0], PHONESTATUS, sizeof (hv1));

	   if (strcmp (hv1, "OK"))
	      return -1;
	}

	if ((p = strchr (hv0, ':')) != NULL)
	{
	   p++;
	   i = atoi (hv0);
	}
	else
	   return -1;

	return i;
}

int DeleteEntry (int fd, int slot)
{
char hv0[1024];
char command[31];
int repeat;

	if (slot < bvon || slot > bbis)
	   return -1;

	/*
	 * Bevor wir einen Slot im Handy loeschen koennen, muessen
	 * wir feststellen, ob der Slot auch tatsaechlich Daten enthaelt.
	 * Deshalb versuchen wir zuerst den Eintrag auszulesen.
	 */
	sprintf (command, "%s%d", RVCALENTRY, slot);
	strcpy (hv0, command);
	strcat (hv0, "\r\n");
	WritePort (fd, hv0, strlen (hv0));
	ReadPort (fd, &hv0[0], command, sizeof (hv0));

	if (!strcmp (hv0, "ERROR"))
	   return -1;

	if (!strcmp (hv0, "OK"))
	   return -3;		// The slot is already empty

	ReadPort (fd, &hv0[0], command, sizeof (hv0));	// We got a data packet

	if (!strcmp (hv0, "ERROR") || !isdigit (hv0[0]))
	   return -1;

	ReadPort (fd, &hv0[0], command, sizeof (hv0));
	
	if (strcmp (hv0, "OK"))
	   return -2;
	   
	/*
	 * An dieser Stelle haben wir festgestellt, das der Slot nicht
	 * leer ist. Im Folgenden setzen wir daher den Loeschbefehl ab.
	 */
	sprintf (command, "%s%d,0", WVCALENTRY, slot);
	strcpy (hv0, command);
	strcat (hv0, "\r\n");
	WritePort (fd, hv0, strlen (hv0));
	repeat = 1;

	do
	{
	   ReadPort (fd, &hv0[0], command, sizeof (hv0));

	   if (!strcmp (hv0, "OK"))
	      repeat = 0;
	   else if (!strcmp (hv0, "ERROR") || !strncmp (hv0, "+CME ERROR", 10))
	      return -1;
	}
	while (repeat);

	return slot;
}

char *GetLinePort (int fd)
{
	return tty_read (fd, "_");
}

void CloseHandy (int fd)
{
	close (fd);
	printf ("Schnittstelle wurde geschlossen.\n");
}

int SendData (int fd, int slot, char *data)
{
char *buf1, *buf2, *p;
char ch[4], hv0[1024], command[31];
int len, i;
int flag;

	if (slot < bvon || slot > bbis)
	   return -1;

	/*
	 * Als erstes muessen wir den String in ein spezielles Format
	 * umwandeln.
	 */
	len = strlen (data);

	if ((buf1 = (char *)malloc (360)) == NULL)
	{
	   fprintf (stderr, "Nicht gengend Speicher um Daten an das Handy zu senden!\n");
	   return -1;
	}

	if ((buf2 = (char *)malloc (360)) == NULL)
	{
	   fprintf (stderr, "Nicht gengend Speicher um Daten an das Handy zu senden!\n");
	   free (buf1);
	   return -1;
	}

	if (Verbose)
	   printf ("Sende Eintrag:\n%s", data);

	memset (buf1, 0, 360);
	memset (buf2, 0, 360);
	p = data;

	/*
	 * Umwandeln des Strings
	 */
	i = 0;

	while (*p && i < (352 * 2))
	{
	   sprintf (ch, "%02X", (int)*p);

	   if (i < 175)
	      strcat (buf1, ch);
	   else
	      strcat (buf2, ch);

	   p++;
	   i++;
	}

	strcat (buf1, "\x1a");
	strcat (buf2, "\x1a");

	if (len > 175)
	{
	   sprintf (command, "%s%d,1,2", WVCALENTRY, slot);
	   flag = 1;
	}
	else
	{
	   sprintf (command, "%s%d,1,1", WVCALENTRY, slot);
	   flag = 0;
	}

	strcpy (hv0, command);
	strcat (hv0, "\r\n");
	WritePort (fd, hv0, strlen (hv0));
	/*
	 * We should get a data request ...
	 */
	ReadPort (fd, &hv0[0], command, sizeof (hv0));

	if (!strcmp (hv0, "ERROR") || !strncmp (hv0, "+CME ERROR", 10))
	{
	   free (buf1);
	   free (buf2);
	   return -1;
	}

	if (hv0[0] != '>')
	{
	   free (buf1);
	   free (buf2);
	   return -2;
	}

	WritePort (fd, buf1, strlen (buf1));
	ReadPort (fd, &hv0[0], buf1, sizeof (hv0));

	if (strcmp (hv0, "OK"))
	{
	   free (buf1);
	   free (buf2);
	   return -1;
	}

	if (flag)
	{
	   sprintf (command, "%s%d,2,2", WVCALENTRY, slot);
	   strcpy (hv0, command);
	   strcat (hv0, "\r\n");
	   WritePort (fd, hv0, strlen (hv0));
	   ReadPort (fd, &hv0[0], command, sizeof (hv0));

	   if (hv0[0] != '>')
	   {
	      free (buf1);
	      free (buf2);
	      return -2;
	   }

	   WritePort (fd, buf2, strlen (buf2));
	   ReadPort (fd, &hv0[0], command, sizeof (hv0));
	}

	free (buf1);
	free (buf2);

	if (strcmp (hv0, "OK"))
	   return -1;

	return 0;
}

/*
 * Ab hier befinden sich alle privaten Funktionen dieses Moduls.
 */
char *ReadPort (int fd, char *buf, char *command, int len)
{
char *value;

	if ((value = tty_read (fd, command)) == NULL)
	{
	   *buf = 0;
	   return NULL;
	}

	strncpy (buf, value, len);
	*(buf+(len-1)) = 0;
	free (value);
	
	if (Verbose)
	   printf ("Received: %s\n", buf);

	return buf;
}

char *tty_read (int fd, char *at_command_p)
{
char *value;
char *temp;
char tempat[BUFSIZ];
int repeat;

	value = NULL;

	do
	{
	   repeat = 0;
	   value = tty_readline (fd);

	   if (value == NULL)
	   {
	      fprintf (stderr, "Error on reading from device: %s\n", strerror (errno));
	      return NULL;
    	   }
	   else if (strlen (value) == 0)
	   {
	      fprintf (stderr, "Reading from the device timed out.\n");
	      free (value);
	      return NULL;
	   }
	   else
	   {
	      if ((temp = index (value, '\r')) || (temp = index (value, '\n')))
	      {
		 //we don't want values 13=<CR>='\r' and 10=<LF>='\n' in our strings
		 memset (temp, 0, 1);
		 //we do not want to return empty strings
		 //we do not accept the at_command_p as answer
		 if (strlen (value) == 0 || !strcmp (value, at_command_p))
		 {
		    free (value);
		    value = NULL;
		    repeat = 1;
		 }
	      }
	   }
	}
	while (repeat);

//	printf ("Received: %s\n", value);

	//PIN handling for all function is done transparently here
	if (!strncmp (value, "+CME ERROR: ", 12) &&
		(strstr (value + 12, "PIN") != NULL ||
		strstr (value + 12, "PUK") != NULL))
	{
	   if (strlen (PIN) == 0)
	   {
	      fprintf (stderr, "%s\nUse the --pin parameter\n", value + 12);
	      free (value);
	      return NULL;
	   }
	   else
	   {
	      memset (tempat, 0, sizeof (tempat));
	      sprintf (tempat, "+CPIN=\"%s\"\r\n", PIN);
	      free (value);
	      WritePort (fd, tempat, strlen (tempat));
	      value = tty_readline (fd);

	      if (strcmp (value, "OK"))
	      {
		 fprintf (stderr, "Error on using pin: %s\n", value + 12);
		 free (value);
		 return NULL;
	      }

	      free (value);
	      WritePort (fd, at_command_p, strlen (at_command_p));
	      value = tty_read (fd, at_command_p);
	   }
	}

	return realloc (value, strlen(value) + 1);
}

/* This function returns a char* that must be freed.
 * If it returns NULL, there was an error. Look at errno for the reason.
 * If the strlen(char*)==0, reading from the device timed out. Retry or abort.
 * (the timeout might only happen when the device was not opened with O_NOBLOCK,
 *  the timeout time refers to the value that was set as c_cc[VTIME])
 */
char *tty_readline (int fd)
{
char* retval;
int retvalsize = 0;
int status;
char buffer;		//buffer the read character for checks
int counter = 0;	//count the read-and-used characters
int repeat = 0;		//tell the inner loop to loop

	retval = NULL;
	do
	{
	   do
	   {
	      status = read (fd, &buffer, 1);

	      if (status == -1)
	      {
		 if (errno == EAGAIN)
		 {
#ifdef WINDBLOWS
            Sleep (1);
#else
		    usleep(1);
#endif
		    repeat=1;
		 }
		 else
		    return NULL;
	      }
	      else if (status == 0)
		 return strdup ("");
	      else
		 repeat = 0;
	   }
	   while (repeat);

	   // allocate space on stack
	   if (counter >= retvalsize)
	   {
	      retvalsize += BUFSIZ;
	      retval = realloc (retval, retvalsize + 1);
	      memset (retval + counter, 0, retvalsize - counter + 1);
	   }

	   if (buffer==0)
	   {
	      /* fix the @=0x00 problem here :-(
	       * we simply add the eigth bit, so when processing:
	       * only look at the last 7 bits (char&127)
	       */
	      retval[counter++] = 128;
	   }
	   else
	   {
	      retval[counter++] = buffer;
	   }
	}
	while (buffer != '\n' && !(retval[0] == '>' && counter > 1));

	/* There are two types of possible answers:
	 * "><space>" (2 character) for data input requests
	 * "....<CR><LF>" for all other things (even empty)
	 */
	return realloc (retval, strlen (retval) + 1);
}

int WritePort (int fd, const char *data, size_t count)
{
int status = 0;
int instatus;

	do
	{
	   instatus = write (fd, &data[status], count - status);

	   if (instatus != -1)
	      status += instatus;
	   else if (status == -1 && errno == EAGAIN)
#ifdef WINDBLOWS
          Sleep (1);
#else
	      usleep(1);
#endif
	}
	while ((status == -1 && errno == EAGAIN) || status < count);
	
	if (Verbose)
	   printf ("Send: %s", data);

	return status;
}

char *HTrim (char *str)
{
char *p1, *p2;

	p1 = p2 = str;

	while (*p1 && *p1 == '\xff')
	   p1++;

	while (*p1)
	{
	   *p2 = *p1;
	   p1++;
	   p2++;
	}

	*p2 = 0;
	return str;
}

#ifdef WINDBLOWS
unsigned long tty_speed (int newttyspeed)
#else
speed_t tty_speed (int newttyspeed)
#endif
{
#ifdef WINDBLOWS
    return (unsigned long)newttyspeed;
#else
	switch(newttyspeed)
	{
	   case 9600:
	      return B9600;
	   break;

	   case 19200:
	      return B19200;
	   break;

	   case 38400:
	      return B38400;
	   break;
#ifdef B57600
	   case 57600:
	      return B57600;
	   break;
#endif
#ifdef B115200
	   case 115200:
	      return B115200;
	   break;
#endif
	   default:
	      fprintf(stderr, "No valid baudrate defined. Using compiled in value...\n");
	      return tty_speed(19200);
	   break;
	}
#endif    // WINDBLOWS
}
